
  
import numpy as np
def relu(inputs):
    return np.maximum(inputs, 0)
def delRelu(rel):
    return int(rel>=0)



def delReluFromRelu(rel):

    return np.vectorize(delRelu)(rel)


# def sigmoid()
def sigmoid(a):
    return 1/(1+np.exp(-a))


def delSigmoid(a):
    sig=sigmoid(a)
    return sig*(1-sig)


def delSigFromSig(sig):

    return sig*(1-sig)


def tanh(a):
    return np.tanh(a)

def delTanhfromTanh(ta):
    return 1-ta*ta

def softmax(inputs):
    exp = np.exp(inputs)
    return exp/np.sum(exp, axis = 1, keepdims = True)

def cross_entropy(inputs, y):
    indices = np.argmax(y, axis = 1).astype(int)
    probability = inputs[np.arange(len(inputs)), indices] 
    log = np.log(probability)
    loss = -1.0 * np.sum(log) / len(log)
    return loss

#mean squared error
def MSE(inputs,y):
    return np.sum((inputs-y)**2)/len(y)

# L2 regularization
def L2_regularization(la, weight):
    return np.sum([0.5 * la * np.sum(w*w) for w in weight])



import numpy as np 
import pickle
# import function as function
import sidSimpleNN.function as function
import time
#import matplotlib.pyplot as plt



class Network:

    def __init__(self, 
                 num_nodes_in_layers, batch=1,epochs=3,learning_rate=0.001, weights_file=None,name='net1',chosenActivation='relu',chosenLoss='CE'
                 ):

        self.num_nodes_in_layers = num_nodes_in_layers


        #include input,hidden,output
        self.no_of_layers=len(num_nodes_in_layers)

        self.batch = batch
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.weights_file = weights_file

        self.w=[np.random.normal(0, 1, [self.num_nodes_in_layers[i], self.num_nodes_in_layers[i+1]]) for i in range(self.no_of_layers-1)]
        self.b=[np.zeros((1, self.num_nodes_in_layers[i])) for i in range(1,self.no_of_layers)]



        self.loss = []

        self.name=name

        self.chosenActivation=chosenActivation
        self.chosenLoss=chosenLoss

        self.aA={
        'relu':[function.relu,function.delReluFromRelu],'sigmoid':[function.sigmoid,function.delSigFromSig],'tanh':[function.tanh,function.delTanhfromTanh]
        }
        self.lF={
        'MSE':function.MSE,'CE':function.cross_entropy
        }
        self.calcLoss=True
        self.applySoftmax=True
        self.applyRegularization=True
        self.regularizationConst=0.01



        #for batch

        self.otherImplementation=True

        self.changeActivation(self.chosenActivation)

        # self.epochs=1
    def saveWeightsBiasJSON(self):
        import json
        w=[i.tolist() for i in self.w]
        b=[i.tolist() for i in self.b]
        with open(self.name+'WB.json', 'w') as outfile:
            json.dump({'weights':w,'biase':b}, outfile)
    def showModel(self):
        print('layers: ',self.num_nodes_in_layers)
        print('learning rate: ',self.learning_rate)
        print('chosenActivation: ',self.chosenActivation)
        print('epochs: ',self.epochs)

    def changeActivation(self,activ):
        [self.activation,self.fasqwdfzeelActivationFromActivation]=self.aA[activ]

    def save(self):
        filehandler = open(self.name, 'wb') 
        pickle.dump(self, filehandler)

    def load(self):

        filehandler = open(self.name, 'rb') 
        return pickle.load(filehandler)

    def lossGraph(self):
        import matplotlib.pyplot as plt
        ypoints = np.array(self.loss)
        xpoints = np.array([i for i in range(len(self.loss))])

        plt.plot(xpoints, ypoints)
        plt.show()

    def activation(self,a):
        return function.relu(a)
    def delActivationFromActivation(self,a):
        return function.delReluFromRelu(a)

    # function.relu
    def feedforward(self,a,activation=function.relu):

        activation=self.activation
        self.fasqwdfzeop=[]
        self.rasqwdfzeopop=[]
        for w,b in zip(self.w,self.b):
            
            self.fasqwdfzeop.append(np.dot(a, w) + b);
            a = activation(self.fasqwdfzeop[-1])

            self.rasqwdfzeopop.append(a);

        return a

    def train(self, inputs, labels):

        self.changeActivation(self.chosenActivation)
        s=time.time()
        for epoch in range(self.epochs):
            it = 0
            while it < len(inputs):

                ib = inputs[it:it+self.batch]
                lb = labels[it:it+self.batch]
                
                self.feedforward(ib)
                y=function.softmax(self.fasqwdfzeop[-1]) if self.applySoftmax else self.rasqwdfzeopop[-1]

                loss=0
                if self.calcLoss:
                    loss = self.lF[self.chosenLoss](y, lb)
                    if self.applyRegularization:
                        loss += function.L2_regularization(self.regularizationConst, self.w)
                    self.loss.append(loss)
                self.fasqwdfze=[(y - lb) / y.shape[0]]



                for i in range(self.no_of_layers-2):
                    pd=self.fasqwdfze[-1]
                    cd=np.dot(pd, self.w[-1-i].T)
   
                    cd*=self.fasqwdfzeelActivationFromActivation(self.rasqwdfzeopop[-1-i-1])
                    self.fasqwdfze.append(cd)


                if self.otherImplementation:
                    for bn in range(self.batch):
                        self.fasdfz=[np.dot(np.array([self.rasqwdfzeopop[-2][bn]]).T, [self.fasqwdfze[0][bn]])]
                        self.fasdfze=[np.sum([self.fasqwdfze[0][bn]], axis = 0, keepdims = True)]
                        for i in range(self.no_of_layers-2-1):
                            self.fasdfz.append(np.dot(np.array([self.rasqwdfzeopop[-3-i][bn]]).T, [self.fasqwdfze[i+1][bn]]))
                            self.fasdfze.append(np.sum([self.fasqwdfze[i+1][bn]], axis = 0, keepdims = True))
                        self.fasdfz.append(np.dot(np.array([ib[bn]]).T, [self.fasqwdfze[-1][bn]]))
                        self.fasdfze.append(np.sum([self.fasqwdfze[-1][bn]], axis = 0, keepdims = True))


                else:
                    self.fasdfz=[np.dot(self.rasqwdfzeopop[-2].T, self.fasqwdfze[0])]
                    self.fasdfze=[np.sum(self.fasqwdfze[0], axis = 0, keepdims = True)]
                    for i in range(self.no_of_layers-2-1):
                        self.fasdfz.append(np.dot(self.fasqwdfzeop[-3-i].T, self.fasqwdfze[i+1]))
                        self.fasdfze.append(np.sum(self.fasqwdfze[i+1], axis = 0, keepdims = True))
                    self.fasdfz.append(np.dot(ib.T, self.fasqwdfze[-1]))
                    self.fasdfze.append(np.sum(self.fasqwdfze[-1], axis = 0, keepdims = True))

                # try:
                for i in range(self.no_of_layers-1):
                    if self.applyRegularization:
                        self.fasdfz[i]+=self.regularizationConst*self.w[-1-i]
                        self.fasdfze[i]+=self.regularizationConst*self.b[-1-i]
                    self.w[-1-i]-=self.learning_rate * self.fasdfz[i]
                    self.b[-1-i]-=self.learning_rate * self.fasdfze[i]

                


                print('=== Epoch: {:d}/{:d}\titeration:{:d}\tLoss: {:.2f} ==='.format(epoch+1, self.epochs, it+1, loss))
                it += self.batch
            # break

        print('=== Time (Training) : {} sec for epoch : {} of input size : {} ===='.format(time.time()-s,self.epochs,len(inputs)))

    def test(self, i, l):

        self.changeActivation(self.chosenActivation)

        s=time.time()
        self.feedforward(i)
        sc=self.fasqwdfzeop[-1]
        pr=function.softmax(sc)
        a = float(np.sum(np.argmax(pr, 1) == l)) / float(len(l))
        print('Test accuracy: {:.2f}%'.format(a*100))
        print('=== Time (Testing) : {} sec ,of input size : {} ===='.format(time.time()-s,len(i)))

        message='Test accuracy: {:.2f}%'.format(a*100)+'\n'+'=== Time (Testing) : {} sec ,of input size : {} ===='.format(time.time()-s,len(i))


        return message




import numpy as np

import mnist
import sidSimpleNN.myNN as myNN 
# import myNN as myNN 
# import myNN 

def run():

	# load data
	num_classes = 10
	train_images = mnist.train_images() #[60000, 28, 28]
	train_labels = mnist.train_labels()
	test_images = mnist.test_images()
	test_labels = mnist.test_labels()

	# print("Training...")

	# # data processing
	X_train = train_images.reshape(train_images.shape[0], train_images.shape[1]*train_images.shape[2]).astype('float32') #flatten 28x28 to 784x1 vectors, [60000, 784]
	x_train = X_train / 255 #normalization
	y_train = np.eye(num_classes)[train_labels] #convert label to one-hot

	X_test = test_images.reshape(test_images.shape[0], test_images.shape[1]*test_images.shape[2]).astype('float32') #flatten 28x28 to 784x1 vectors, [60000, 784]
	x_test = X_test / 255 #normalization
	y_test = test_labels

	np.random.seed(1)

	net = myNN.Network(
                 num_nodes_in_layers = [784, 20, 10], 
                 batch = 1,
                 epochs = 6,
                 learning_rate = 0.001, 
                 weights_file=None,
                 chosenActivation='tanh',
                 chosenLoss='MSE',
                 # name='tempnet3'
                 # weightsAndBias_file = 'digitRecog',
                 # includeWeightsBias=True

             )

	print('before training , testing with test dataset')
	# net.chosenLoss='CE'
	# net.applySoftmax=True
	# net.applyRegularization=False
	# net.regularizationConst=0.01
	# net.calcLoss=False
	# net.saveWeightsBiasJSON()
	message=net.test(x_test, y_test)


	net.train(x_train, y_train)
	net.save()
	net.lossGraph()
	print('before training , testing with test dataset')
	print(message)

	print("after training")
	net.test(x_test, y_test)
	net.showModel()

	# net.applyRegularization
	# net.regularizationConst=0.0001
	# net.applySoftmax=False
	# np.set_printoptions(suppress=True)
	# net.feedforward(x_train[0])
	# z=net.fasqwdfzeop
	#activation applied
	# a=net.rasqwdfzeopop
	# print(z)
	# print(a)
	# import function
	
	# print(function.tanh([[-26.33910613 ,  7.56055857, -47.45400574,   1.65570075,  -5.44114857,
	#  -23.85436946,  11.8361229,   16.61606579, -43.88570782, -27.62500334]]))

	# print(np.array(list(map(np.tanh,z))))

if __name__=='__main__':
	run()





from setuptools import setup, find_packages

with open('README.md','r') as readme_file:
    README = readme_file.read()

# with open('HISTORY.md') as history_file:
#     HISTORY = history_file.read()

setup_args = dict(
    name='sidSimpleNN',
    version='0.2.8',
    description='Multilayered Perceptron NN light weight, make, train, test models',
    # long_description_content_type="text",
    # long_description='this is simmple NN library\n\nfrom sidSimpleNN import mainTrain\n mainTrain.run()\nif data,mnist folder dont get made automatically then manualy download from MNIST dataset and place in datafolder ',
    
    long_description=README,
    long_description_content_type='text/markdown',
    # long_description=README,
    license='MIT',
    packages=find_packages(),
    author='sid007',
    author_email='sid700710@gmail.com',
    keywords=['sidSimpleNN'],
    url='',
    download_url='https://pypi.org/project/sidSimpleNN/'
)

# install_requires = [
#     'pickle','numpy','matplotlib'
# ]


if __name__ == '__main__':
    setup(**setup_args)